class NotFoundError(Exception):
    """Raised when not found."""
    pass

class ElasticsearchError(Exception):
    """Raised elasticsearch Error."""
    pass

class DocPathsWithoutValuesAggregationError(Exception):
    """Raised when the user defines aggregations with include_values=False and include_doc_paths=True"""
    pass

class KeyFieldEqualsValueFieldError(Exception):
    """Raised when the user tries to use the same value for key_field and value_field while aggregating over facts in facts_abstract."""
    pass

class InvalidKeyFieldError(Exception):
    """Raised when the user tries to use inavlid key_field while aggregating over facts in facts_abstract."""
    pass

class InvalidValueFieldError(Exception):
    """Raised when the user tries to use invalid value_field while aggregating over facts in facts_abstract."""
    pass



from typing import List, Optional

import elasticsearch
from elasticsearch_dsl import Search
from elasticsearch_dsl.query import MoreLikeThis

from texta_elastic.core import ElasticCore, elastic_connection
from texta_elastic.settings import TEXTA_TAGS_KEY


ES_SCROLL_SIZE = 500
EMPTY_QUERY = {"query": {"match_all": {}}}


class ElasticSearcher:
    """
    Everything related to performing searches in Elasticsearch
    """
    OUT_RAW = 'raw'  # Gives you a list of documents that belong to the scroll batch.
    OUT_DOC = 'doc'  # Gives you only the _source content without index and id metadata.
    OUT_TEXT = 'text'
    OUT_TEXT_WITH_ID = "text_with_id"
    OUT_DOC_WITH_ID = 'doc_with_id'
    OUT_DOC_WITH_TOTAL_HL_AGGS = 'doc_with_total_hl_aggs'
    OUT_META = 'out_meta'


    def __init__(self,
                 field_data=[],
                 indices=[],
                 query=EMPTY_QUERY,
                 scroll_size=ES_SCROLL_SIZE,
                 output=OUT_DOC,
                 callback_progress=None,
                 scroll_limit=None,
                 ignore_ids=set(),
                 text_processor=None,
                 score_threshold=0.0,
                 filter_callback=None,
                 timeout='10m',
                 flatten=True,
                 scroll_timeout: str = None,
                 elastic_core: Optional[ElasticCore] = None):
        """
        :param field_data: List of fields names you want returned from Elasticsearch. Specify the fields if you only need a single field to save on bandwidth and transfer speeds.
        :param indices: List of index names from which to pull data.
        :param query: Query for Elasticsearch.
        :param scroll_size: How many items should be pulled with each scroll request.
        :param output: Constant for determine document output.
        :param flatten: Whether to flatten output document or not.
        :param callback_progress: Function to call after each successful scroll request.
        :param scroll_limit: Number of maximum documents that are returned from the scrolling process.
        :param ignore_ids: Iterable of Elasticsearch document ID's which are not returned.
        :param text_processor: Text processor object to... process text.
        :param score_threshold: Filters out documents which score value don't exceed the given limit.
        :param filter_callback: Function body that takes a document dictionary as input and returns in, used for filtering fields.
        :param timeout: Time in string for how long to wait for an Elasticsearch request.
        :param scroll_timeout: Time in string for how long to keep scroll context in memory, if not explicitly set, defaults to request timeout.
        :param elastic_core: Instance of ElasticCore to use instead of one created with default parameters.
        """
        self.core = elastic_core or self._get_core()
        self.field_data: List[str] = field_data
        self.indices = indices
        self.query = query
        self.scroll_size = scroll_size
        self.scroll_limit = scroll_limit
        self.score_threshold = score_threshold
        self.ignore_ids = ignore_ids
        self.output = output
        self.flatten = flatten
        self.callback_progress = callback_progress
        self.text_processor = text_processor
        self.timeout = timeout
        self.filter_callback = filter_callback
        self.scroll_timeout = scroll_timeout or timeout

        if self.callback_progress:
            total_elements = self.count()
            if scroll_limit and scroll_limit < total_elements:
                total_elements = scroll_limit
            callback_progress.set_total(total_elements)


    def _get_core(self):
        return ElasticCore()


    def __iter__(self):
        """
        Iterator for iterating through scroll
        """
        return self.scroll()


    def more_like_this(self, mlt_fields: List[str], like, exclude=[], min_term_freq=1, max_query_terms=12, min_doc_freq=5, min_word_length=0, max_word_length=0, stop_words=[], size=10, include_meta=False, indices: str = None, flatten: bool = False):
        """

        Args:
            indices: Coma-separated string of the indices you wish to use.
            mlt_fields: List of strings of the fields you wish to use for analyzation.
            like: Can either be a text field or a list of document metas which is used as a baseline for fetching similar documents.
            exclude: List of document ids that should be ignored.
            min_term_freq: The minimum term frequency below which the terms will be ignored from the input document.
            max_query_terms: The maximum number of query terms that will be selected. Increasing this value gives greater accuracy at the expense of query execution speed.
            min_doc_freq: The minimum document frequency below which the terms will be ignored from the input document.
            min_word_length: The minimum word length below which the terms will be ignored.
            max_word_length: The maximum word length above which the terms will be ignored.
            stop_words: An array of stop words. Any word in this set is considered "uninteresting" and ignored.
            include_meta: Whether to add the documents meta information (id, index, doctype) into the returning set of documents.
            size: How many documents to return with the end result.
            flatten: Whether to flatten nested fields.
        Returns: List of Elasticsearch documents.

        """
        indices = indices if indices else ",".join(self.indices)
        s = Search(using=self.core.es, index=indices)
        mlt = MoreLikeThis(like=like, fields=mlt_fields, min_term_freq=min_term_freq, max_query_terms=max_query_terms, min_doc_freq=min_doc_freq, min_word_length=min_word_length, max_word_length=max_word_length, stop_words=stop_words)
        s = s.query(mlt).exclude("ids", values=exclude)
        s = s.extra(size=size)
        if include_meta:
            response = []
            for hit in s.execute():
                item = {
                    "_id": hit.meta.id,
                    "_index": hit.meta.index,
                    "_type": getattr(hit.meta, "doc_type", "_doc"),
                    "_source": self.core.flatten(hit.to_dict()) if flatten else hit.to_dict()
                }
                response.append(item)
            return response
        else:
            response = [self.core.flatten(hit.to_dict()) if flatten else hit.to_dict() for hit in s.execute()]
            return response


    def update_query(self, query):
        self.query = query


    def update_field_data(self, field_data):
        self.field_data = field_data


    def _check_for_field_data(self, dict_key: str) -> bool:
        """
        Helper function for _parse_doc to ensure that the root of nested fields
        and that the leaf of the nested field aren't filtered away by it.

        Given the input: {"text": {"content": "arrividerci"}}, it should not filter out fields
        if the given field in self.field_data is either text.content or just text.
        """
        root = dict_key.split(".")[0]
        is_full_nested_path = dict_key in self.field_data
        only_root_key_is_in_nested_path = root in self.field_data
        return any([is_full_nested_path, only_root_key_is_in_nested_path])


    def _default_filter(self, doc: dict):
        if self.field_data:
            # Splitting k by a dot allows us to also include nested objects which would otherwise be filtered out
            # as they won't match the fields names: so id a document has {"text": {"content": ""}} it will appear as
            # text.content which would otherwise fail the check for just "text".
            parsed_doc = {k: v for k, v in doc.items() if self._check_for_field_data(k)}
            return parsed_doc
        return doc


    def _parse_doc(self, doc: dict):
        """
        Parses a potentially nested dictionary into a flat one with it's keys delimited by a dot.
        By default, it also only keeps the fields that have been included inside self.field_data.
        """
        parsed_doc, _, _ = self._flatten_doc(doc)
        parsed_doc = self.filter_callback(parsed_doc) if self.filter_callback else self._default_filter(parsed_doc)
        return parsed_doc


    def _parse_doc_with_highlight(self, doc):
        """
        Parses Elasticsearch hit into something nicer, includes the highlight field
        """
        parsed_doc, _, highlight = self._flatten_doc(doc)
        return {'highlight': highlight, 'doc': parsed_doc}


    def _flatten_doc(self, doc: dict):
        """
        Flattens an Elasticsearch document by changing it's nested structures into a flat one
        where the paths of the keys are made into a dot delimited one.

        {"text": {"content": "arrividerci"}} -> {"text.content": "arrividerci"}

        Returns the flattened source, the name of the index and the highlight portion.

        """
        index = doc['_index']
        highlight = doc['highlight'] if 'highlight' in doc else {}
        doc = doc['_source']
        if self.flatten:
            doc = self.core.flatten(doc)
        return doc, index, highlight


    def _decode_doc(self, doc, field_path=None):
        decoded_text = doc['_source']
        if field_path:
            # decode if field path known
            for k in field_path.split('.'):
                # get nested fields encoded as: 'field.sub_field'
                try:
                    decoded_text = decoded_text[k]
                except:
                    decoded_text = ""
        else:
            pass
        return decoded_text


    def count(self) -> int:
        try:
            count = self.core.es.count(index=self.indices, body=self.query)
            return count["count"]
        except elasticsearch.NotFoundError:
            return 0


    def search(self, size=10):
        # by default return all fields
        source_fields = True
        if self.output == self.OUT_META:
            source_fields = False
        # In case size/from is included in query in pagination, don't overwrite it by passing the size parameter
        if 'size' in self.query:
            response = self.core.es.search(index=self.indices, body=self.query, timeout=self.timeout, _source=source_fields)
        else:
            response = self.core.es.search(index=self.indices, body=self.query, size=size, timeout=self.timeout, _source=source_fields)
        if self.output == self.OUT_DOC:
            hits = [self._parse_doc(doc) for doc in response['hits']['hits']]
            return hits
        if self.output == self.OUT_DOC_WITH_ID:
            for hit in response['hits']['hits']:
                parsed_doc, _, highlight = self._flatten_doc(hit)
                hit['highlight'] = highlight
                hit['_source'] = parsed_doc
            return response
        if self.output == self.OUT_DOC_WITH_TOTAL_HL_AGGS:
            return {
                'count': response['hits']['total'],
                'aggs': response['aggregations'] if 'aggregations' in response else {},
                'results': [self._parse_doc_with_highlight(doc) for doc in response['hits']['hits']]
            }

        else:
            return response


    def random_documents(self, size=10):
        """
        Returns n random documents, where n=size.
        """
        random_query = {
            "query": {
                "function_score": {
                    "query": self.query["query"],
                    "functions": [{"random_score": {}}]
                }
            }
        }
        response = self.core.es.search(index=self.indices, body=random_query, size=size)
        if self.output == self.OUT_DOC:
            return [self._parse_doc(doc) for doc in response['hits']['hits']]
        else:
            return response


    def resolve_field_data(self) -> List[str]:
        """
        Ensure that regardless of the fields inserted, it also always returns the facts
        to avoid problems with other applications.
        """
        if not self.field_data:
            return ["*"]
        elif self.field_data and TEXTA_TAGS_KEY not in self.field_data:
            return self.field_data + [TEXTA_TAGS_KEY]
        elif self.field_data and TEXTA_TAGS_KEY in self.field_data:
            return self.field_data
        else:
            return ["*"]


    # batch search makes an inital search, and then keeps pulling batches of results, until none are left.
    @elastic_connection
    def scroll(self):
        scroll_id = None

        try:

            # Zero-out the progress in case the same ElasticSearch instance is used twice while iterating through the dataset
            if self.callback_progress:
                self.callback_progress.update_view(0)
                self.callback_progress.n_count = 0

            field_data = self.resolve_field_data()
            if self.output == self.OUT_META:
                page = self.core.es.search(index=self.indices, body=self.query, scroll=self.scroll_timeout, _source_excludes=["*"], size=self.scroll_size)
            else:
                page = self.core.es.search(index=self.indices, body=self.query, scroll=self.scroll_timeout, _source=field_data, size=self.scroll_size)

            scroll_id = page['_scroll_id']
            current_page = 0
            # set page size
            page_size = len(page['hits']['hits'])
            num_scrolled = 0
            # set score threshold
            if page['hits']['max_score']:
                lowest_allowed_score = page['hits']['max_score'] * self.score_threshold
            else:
                lowest_allowed_score = self.score_threshold
            # set scroll break default
            scroll_break = False
            # iterate through scroll
            while page_size > 0 and scroll_break is False:
                # process output
                if self.output in (self.OUT_DOC, self.OUT_DOC_WITH_ID, self.OUT_TEXT, self.OUT_TEXT_WITH_ID, self.OUT_META):
                    if self.callback_progress:
                        self.callback_progress.update(page_size)
                    for hit in page['hits']['hits']:
                        # if scroll limit reached, break the scroll
                        if self.scroll_limit and num_scrolled >= self.scroll_limit:
                            scroll_break = True
                            break
                        # if score too low, break scroll
                        elif hit['_score'] < lowest_allowed_score:
                            scroll_break = True
                            break
                        if hit['_id'] not in self.ignore_ids:
                            num_scrolled += 1
                            parsed_doc = self._parse_doc(hit)
                            if self.output == self.OUT_META:
                                yield hit
                            elif self.output == self.OUT_TEXT:
                                for field in parsed_doc.values():
                                    if self.text_processor:
                                        field = self.text_processor.process(field)
                                        for text in field:
                                            yield " ".join(text)
                                    else:
                                        yield field

                            elif self.output == self.OUT_TEXT_WITH_ID:
                                document = {}
                                for key, value in parsed_doc.items():
                                    if key in self.field_data:
                                        processed_field = self.text_processor.process(value)
                                        document[key] = processed_field
                                yield hit["_id"], document

                            elif self.output in (self.OUT_DOC, self.OUT_DOC_WITH_ID):
                                if self.text_processor:
                                    parsed_doc = {k: '\n'.join(self.text_processor.process(v)[0]) for k, v in parsed_doc.items()}

                                if self.output == self.OUT_DOC_WITH_ID:
                                    parsed_doc['_id'] = hit['_id']
                                yield parsed_doc

                # return raw hit
                elif self.output == self.OUT_RAW:

                    if self.callback_progress:
                        self.callback_progress.update(page_size)

                    # filter page by score
                    page = [doc for doc in page["hits"]["hits"] if doc['_score'] >= lowest_allowed_score]

                    # if score too low, break scroll
                    if not page:
                        scroll_break = True
                        break
                    # filter by ignored ids
                    page = [doc for doc in page if doc['_id'] not in self.ignore_ids]
                    if page:
                        num_scrolled += len(page)
                        yield page

                # get new page
                page = self.core.es.scroll(scroll_id=scroll_id, scroll=self.scroll_timeout)
                scroll_id = page['_scroll_id']
                page_size = len(page['hits']['hits'])
                current_page += 1

            if scroll_id:
                self.core.es.clear_scroll(body={'scroll_id': scroll_id})
        except Exception as e:
            if scroll_id:
                self.core.es.clear_scroll(body={'scroll_id': scroll_id})
            raise e



from typing import List
import elasticsearch
from elasticsearch_dsl import A, Q, Search

from texta_elastic.settings import TEXTA_ES_URL


class SpamDetector:
    FILTERED_FIELD_TYPES = ["date", "fact"]


    def __init__(self, indices: List[str]):
        self.indices = indices
        self.es_url = TEXTA_ES_URL
        self.es = elasticsearch.Elasticsearch(self.es_url, timeout=60)
        self.search = Search(using=self.es, index=self.indices)
        self.result = None  # Placeholder for the query results.


    def set_hits_to_zero(self):
        """Set the amount of hits returned to zero, saves network traffic."""
        self.search = self.search.extra(size=0)


    def filter_fields(self, relevant_field_paths: List[str], all_field_info: List[dict]) -> List[dict]:
        """

        :param relevant_field_paths: Name of the fields we want to keep.
        :param all_field_info: All fields we are reducing.
        :return:
        """
        without_wrong_types = list(filter(lambda x: x["type"] not in self.FILTERED_FIELD_TYPES, all_field_info))
        only_relevant_fields = list(filter(lambda x: x["path"] in relevant_field_paths, without_wrong_types))
        return only_relevant_fields


    def apply_date_filter_aggregation(self, **kwargs):
        """
        By default aggregations are applied to the whole dataset,
        to limit it you need to apply a filter aggregation with the needed
        query.
        :param kwargs:
        :return:
        """
        date_field = kwargs["date_field"]
        from_date = kwargs["from_date"]
        to_date = kwargs["to_date"]

        # You need a filter aggregation to limit the documents of other aggregations.
        date_filter = A("filter", Q("range", **{date_field: {'gte': from_date, 'lte': to_date}}))
        self.search.aggs.bucket("date_filter", date_filter)


    def apply_terms_aggregation(self, **kwargs):
        """
        Most frequent items aggregation that is being attached
        to the premade date filter aggregation.
        :param kwargs:
        :return:
        """
        target_field = kwargs["target_field"]
        aggregation_size = kwargs["aggregation_size"]
        min_doc_count = kwargs["min_doc_count"]
        # Frequent items aggregation to find posts with the same values.
        term_filter = A("terms", field="{}.keyword".format(target_field), size=aggregation_size, min_doc_count=min_doc_count)
        # Applying the previous two pre-made aggregations into the query.
        self.search.aggs["date_filter"].bucket("spam", term_filter)


    def apply_nested_terms_aggregations(self, **kwargs):
        """
        Every field send to us is being turned into a most frequent items
        aggregation that's being attached to the previous one as a nested
        aggregation to find common patterns.
        :param kwargs:
        :return:
        """
        min_doc_count = kwargs["min_doc_count"]
        aggregation_size = kwargs["aggregation_size"]
        common_feature_fields = kwargs["common_feature_fields"]

        for field in common_feature_fields:
            if field["type"] == "text":
                elastic_field_name = "{}.keyword".format(field["path"])
                self.search.aggs["date_filter"]["spam"].bucket(
                    field["path"],
                    "terms",
                    field=elastic_field_name,
                    size=aggregation_size,
                    min_doc_count=min_doc_count
                )
            else:
                elastic_field_name = "{}".format(field["path"])
                self.search.aggs["date_filter"]["spam"].bucket(
                    field["path"],
                    "terms",
                    field=elastic_field_name,
                    size=aggregation_size,
                    min_doc_count=min_doc_count
                )


    def execute_query(self):
        self.result = self.search.execute()


    def format_elastic_response(self):
        """
        If it looks messy and unreasonable, it's because it's
        converting one data structure into a different format.
        Do not tread this unless you want to change the outcome.
        """
        ignored_keys = ["doc_count", "key"]
        container = []

        for hit in self.result.aggs.date_filter.spam.buckets:
            coocurrances = []
            frequent_texts = {"value": hit.key, "count": hit.doc_count}
            for field_name in dir(hit):
                if field_name not in ignored_keys:
                    occurrence = getattr(hit, field_name)
                    for agg in occurrence.buckets:
                        coocurrances.append({"value": agg.key, "count": agg.doc_count, "field": field_name})

            container.append({**frequent_texts, "co-occurances": coocurrances})

        self.result = container


    def get_spam_content(self, **kwargs) -> List[dict]:
        """
        Default values of the serializer should be hit into this.
        """
        self.set_hits_to_zero()
        self.apply_date_filter_aggregation(**kwargs)
        self.apply_terms_aggregation(**kwargs)
        self.apply_nested_terms_aggregations(**kwargs)
        self.execute_query()
        self.format_elastic_response()
        return self.result



import logging

import elasticsearch

from .settings import ERROR_LOGGER
from texta_elastic.exceptions import NotFoundError, ElasticsearchError


def elastic_connection(func):
    """
    Decorator for wrapping Elasticsearch functions that are used in views,
    to return a properly formatted error message during connection issues
    instead of the typical HTTP 500 one.
    """


    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        except elasticsearch.exceptions.NotFoundError as e:
            logging.getLogger(ERROR_LOGGER).error(e.info)
            raise NotFoundError(e.info)

        except elasticsearch.exceptions.AuthorizationException as e:
            logging.getLogger(ERROR_LOGGER).warning(e.info)
            error = [error["reason"] for error in e.info["error"]["root_cause"]]
            raise error

        except elasticsearch.exceptions.AuthenticationException as e:
            logging.getLogger(ERROR_LOGGER).warning(e.info)
            raise ElasticsearchError(e.info)

        except elasticsearch.exceptions.TransportError as e:
            logging.getLogger(ERROR_LOGGER).exception(e.info)
            raise ElasticsearchError(e.error)

        except elasticsearch.exceptions.ConnectionTimeout as e:
            logging.getLogger(ERROR_LOGGER).error(e.info)
            raise ElasticsearchError(e.info)


    return func_wrapper



import json
from typing import List

from texta_elastic.settings import TEXTA_TAGS_KEY


INNER_HITS_MAX_SIZE = 100


class Query:

    def __init__(self, operator="must", inner_hits_max_size: int = INNER_HITS_MAX_SIZE):
        self.inner_hits_max_size = inner_hits_max_size
        self.query = {"query": {"bool": {"should": [], "must": [], "must_not": []}}}
        self.operator = operator


    def __str__(self):
        return json.dumps(self.query, ensure_ascii=False)


    def __dict__(self):
        return self.query


    def add_sub_query(self, sub_query):
        self.query["query"]["bool"][self.operator].append(sub_query["query"])


    def add_fact_filter(self, fact_name, fact_value, filter_id=1):
        """
        Adds fact filter to existing query.
        """
        query = {"nested": {
            "path": TEXTA_TAGS_KEY,
            "inner_hits": {
                "name": "fact_filter_{}".format(filter_id),
                "size": self.inner_hits_max_size
            },
            "query": {
                "bool": {
                    "must": [
                        {"match": {f"{TEXTA_TAGS_KEY}.fact": fact_name}},
                        {"match": {f"{TEXTA_TAGS_KEY}.str_val": fact_value}}
                    ]
                }
            }
        }
        }

        self.query["query"]["bool"][self.operator].append(query)


    def add_facts_filter(self, fact_names: List[str], fact_values: List[str] = [], operator: str = "must", filter_id: int = 1):
        """
        Adds facts filter to the query. Fact values are optional, but if added
        len(fact_values) should equal len(fact_names) and fact value in position i
        should correspond to fact name in position i.

        :param fact_names: List of fact names to use for restricting the query.
        :param fact_values: List of fact values (corresponding to the fact names on
                            the same position) to use for restricting the query.
        :param operator: Operator used for joining the facts in the query.
        """
        self.query["query"]["bool"][self.operator].append({"bool": {operator: []}})
        for i, fact_name in enumerate(fact_names):
            new_restriction = {
                "nested": {
                    "path": TEXTA_TAGS_KEY,
                    "inner_hits": {
                        "name": "fact_filter_{}".format(filter_id + i),
                        "size": self.inner_hits_max_size
                    },
                    "query": {
                        "bool": {
                            "must": [
                                {"match": {f"{TEXTA_TAGS_KEY}.fact": fact_name}},
                            ]
                        }
                    }
                }
            }
            if fact_values:
                fact_value = fact_values[i]
                if fact_value:
                    new_restriction["nested"]["query"]["bool"]["must"].append({"match": {f"{TEXTA_TAGS_KEY}.str_val": fact_value}})
            self.query["query"]["bool"]["must"][0]["bool"][operator].append(new_restriction)


    def add_mlt(self, mlt_fields, text):
        """
        Adds More Like This query.
        """
        mlt_query = {"more_like_this": {"fields": mlt_fields, "like": text, "min_term_freq": 1, "max_query_terms": 12}}
        self.query["query"] = mlt_query


    def add_string_filter(self, query_string, match_type="match", fields=None):
        """
        Adds string filter to the query.
        """
        string_matching_query = {"multi_match": {"query": query_string}}
        if fields:
            string_matching_query["multi_match"]["fields"] = fields
        if match_type in ("phrase", "phrase_prefix"):
            string_matching_query["multi_match"]["type"] = match_type
        self.query["query"]["bool"][self.operator].append(string_matching_query)




class SchemaGenerator:
    def _get_text_structure(self):
        txt_structure = {"type": "text",
                         "fields": {
                             "keyword": {
                                 "type": "keyword",
                                 "ignore_above": 256
                             }
                         }}
        return txt_structure

    def _get_long_structure(self):
        long_structure = {"type": "long",
                          "fields": {
                              "keyword": {
                                  "type": "keyword",
                                  "ignore_above": 256
                              }
                          }}
        return long_structure

    def _get_float_structure(self):
        float_structure = {"type": "float",
                           "fields": {
                              "keyword": {
                                  "type": "keyword",
                                  "ignore_above": 256
                              }
                          }}
        return float_structure

    def _get_date_structure(self):
        date_structure = {"type": "date"}
        return date_structure

    def _get_fact_structure(self):
        fact_structure = {"type": "nested",
                          "properties": {
                              "description": {
                                  "type": "text",
                                  "fields": {
                                      "keyword": {
                                          "type": "keyword",
                                          "ignore_above": 256
                                      }
                                  }
                              },
                              "doc_path": {
                                  "type": "keyword"
                              },
                              "fact": {
                                  "type": "keyword"
                              },
                              "num_val": {
                                  "type": "long"
                              },
                              "spans": {
                                  "type": "keyword"
                              },
                              "str_val": {
                                  "type": "keyword"
                              },
                              "sent_index": {
                                  "type": "long"
                              }
                          }}

        return fact_structure

    def _get_texta_meta_structure(self):
        texta_meta_structure = {"type": "object",
                          "properties": {
                              "document_uuid": {
                                  "type": "text"
                              }
                          }}

        return texta_meta_structure

    def get_nested_structure(self, nested_fields):
        nested = {}
        for field in nested_fields:
            tokens = field.split('.')
            for i, token in enumerate(tokens):
                if i == 0:
                    if token not in nested:
                        nested[token] = {'properties': {}}
                    nested_branch = nested
                elif i < len(tokens) - 1:
                    if token not in nested_branch[tokens[i - 1]]['properties']:
                        nested_branch[tokens[i - 1]]['properties'][token] = {'properties': {}}
                    nested_branch = nested_branch[tokens[i - 1]]['properties']
                else:
                    if token not in nested_branch[tokens[i - 1]]['properties']:
                        nested_branch[tokens[i - 1]]['properties'][token] = self._get_text_structure()
                    nested_branch = nested_branch[tokens[i - 1]]['properties']
        return nested

    def init_fields(self, fields):
        keys = ['text', 'date', 'long', 'float','nested', 'texta_facts', 'texta_meta']
        for key in keys:
            if key not in fields:
                fields.update({key: []})
        return fields

    def generate_schema(self, fields, add_facts_mapping=False, add_texta_meta_mapping=False):
        '''
        fields: dict
        '''
        fields = self.init_fields(fields)
        schema = {'properties': {}}
        text_structure = self._get_text_structure()
        long_structure = self._get_long_structure()
        float_structure = self._get_float_structure()
        date_structure = self._get_date_structure()
        fact_structure = self._get_fact_structure()
        texta_meta_structure = self._get_texta_meta_structure()

        if add_facts_mapping:
            schema['properties']['texta_facts'] = fact_structure

        if add_texta_meta_mapping:
            schema['properties']['texta_meta'] = texta_meta_structure

        for field in fields['text']:
            schema['properties'][field] = text_structure
        for field in fields['long']:
            schema['properties'][field] = long_structure
        for field in fields['float']:
            schema['properties'][field] = long_structure
        for field in fields['date']:
            schema['properties'][field] = date_structure
        for field in fields['texta_facts']:
            schema['properties']['texta_facts'] = fact_structure
        for field in fields['texta_meta']:
            schema['properties']['texta_meta'] = texta_meta_structure
        if fields['nested']:
            nested_mapping = self.get_nested_structure(fields['nested'])
            schema['properties'].update(nested_mapping)

        return schema



import logging
from datetime import datetime
from typing import List, Tuple
import sys


# MutableMapping was moved under collection.abc after Python 10.
if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping

import elasticsearch
import elasticsearch_dsl
import requests
from elasticsearch import Elasticsearch
from elasticsearch_dsl import Date, Keyword, Long, Mapping, Nested, Object, Q, Text

from texta_elastic.decorators import elastic_connection
from texta_elastic.exceptions import NotFoundError
from texta_elastic.settings import ES_CONNECTION_PARAMETERS, TEXTA_ANNOTATOR_KEY, TEXTA_ES_PASSWORD, TEXTA_ES_PREFIX, TEXTA_ES_URL, TEXTA_ES_USERNAME, TEXTA_TAGS_KEY


class ElasticCore:
    """
        Class for holding most general settings and Elasticsearch object itself
    """


    def __init__(
            self,
            ES_URL=TEXTA_ES_URL,
            es_prefix=TEXTA_ES_PREFIX,
            es_username=TEXTA_ES_USERNAME,
            es_password=TEXTA_ES_PASSWORD,
            connection_parameters: dict = ES_CONNECTION_PARAMETERS,
            check_connection=True
    ):
        self.logger = logging.getLogger("elastic_core")

        self.es_url = ES_URL
        self.es_prefix = es_prefix
        self.es_username = es_username
        self.es_password = es_password
        self.connection_parameters = connection_parameters

        self.texta_reserved = ['texta_facts']

        # do not connect if asked
        if check_connection:
            self.connection = self._check_connection()
        else:
            self.connection = False
        self.es = self._create_client_interface()

        self.logger.debug(f"Created ElasticCore instance for cluster: {self.es_url}")


    def _create_client_interface(self):
        """
        Support using multiple hosts by splitting a coma-separated ES_URL.
        Having empty strings for auth is safe and does nothing if ES isn't configured for users.
        For safety's sake we remove all connection parameters with None (default if not configured in env),
        and then throw the existing ones with dictionary unpacking as per the Urllib3HttpConnection class.
        """
        if self.connection:
            list_of_hosts = self.es_url.split(",")
            existing_connection_parameters = dict((key, value) for key, value in self.connection_parameters.items() if value is not None)
            client = Elasticsearch(list_of_hosts, http_auth=(self.es_username, self.es_password), **existing_connection_parameters)
            return client


    def _check_connection(self):
        try:
            response = requests.get(self.es_url, timeout=3)
            if response.ok:
                return True
            else:
                return False
        except Exception as e:
            raise NotFoundError(f"Error connecting to Elasticsearch: '{self.es_url}'! Do you have the right URL configured?")


    @staticmethod
    def check_for_security_xpack(elastic_core=None) -> bool:
        """
        Checks whether the Elasticsearch Security X-Pack module is in use
        with its SSL and user authentication support.
        """
        try:
            ec = elastic_core or ElasticCore()
            info = ec.es.xpack.info(categories="features")
            available = info["features"]["security"]["available"]
            return available
        except elasticsearch.exceptions.RequestError as e:
            # When XPACK is not installed the xpack module will lose the info function,
            # which will then create an error, return False in that case.
            return False


    @elastic_connection
    def create_index(self, index, body=None, ignore=(400, 404)):
        return self.es.indices.create(index=index, body=body, include_type_name=True, ignore=ignore)


    @elastic_connection
    def delete_index(self, index: str, ignore=(400, 404)) -> dict:
        """
        :param ignore: Which status code responses should not throw an exception.
        :param index: Name of single index to delete or a coma separated list of indices to deleted.
        :returns: Either {'acknowledged': True} or a detailed error response.
        """
        return self.es.indices.delete(index=index, ignore=ignore)


    @elastic_connection
    def get_index_creation_date(self, index) -> str:
        """
        Returns the date an index was created at in iso-format.
        """
        es_index_settings = self.get_index_settings(index)
        utc_time = datetime.utcfromtimestamp(0).isoformat()
        if str(index) in es_index_settings:
            unix_timestamp = int(es_index_settings[str(index)]['settings']['index']['creation_date']) / 1000
            utc_time = datetime.utcfromtimestamp(unix_timestamp).isoformat()
        return utc_time


    @elastic_connection
    def get_index_settings(self, index):
        return self.es.indices.get_settings(index=index)


    @elastic_connection
    def get_settings(self):
        return self.es.indices.get_settings()


    @elastic_connection
    def get_mapping(self, index):
        return self.es.indices.get_mapping(index=index)


    @elastic_connection
    def close_index(self, index: str):
        return self.es.indices.close(index=index, ignore=[400, 404])


    @elastic_connection
    def open_index(self, index: str):
        return self.es.indices.open(index=index, ignore=[400, 404])


    @elastic_connection
    def build_flavor(self):
        return self.es.info()["version"]["build_flavor"]


    @elastic_connection
    def get_version(self) -> Tuple[int, int, int]:
        """
        :return: All three numbers of the Elasticsearch cluster version as integers.
        """
        info = self.es.info()
        version = info["version"]["number"]
        first, second, third = version.split(".")
        return int(first), int(second), int(third)


    @elastic_connection
    def syncher(self):
        """
        Wipe the slate clean and create a new set of Index objects.
        Since we're not using the destroy views, no actual deletion/creation operations
        will be done on the Elasticsearch cluster.

        Put this into a separate function to make use of it.
        """
        from toolkit.elastic.index.models import Index

        opened, closed = self.get_indices()

        # Delete the parts that exist in the toolkit but not in Elasticsearch.
        es_set = {index for index in opened + closed}
        tk_set = {index.name for index in Index.objects.all()}
        for index in tk_set:
            if index not in es_set:
                Index.objects.get(name=index).delete()

        # Create an Index object if it doesn't exist.
        # Ensures that changes Elastic-side on the open/closed state are forcefully updated.
        es_settings = self.get_settings()
        utc_time = datetime.utcfromtimestamp(0).isoformat()
        for index in opened:
            index, is_created = Index.objects.get_or_create(name=index)
            if str(index) in es_settings:
                unix_timestamp = int(es_settings[str(index)]['settings']['index']['creation_date']) / 1000
                utc_time = datetime.utcfromtimestamp(unix_timestamp).isoformat()
            index.created_at = utc_time
            index.save()
            if not index.is_open:
                index.is_open = True
                index.save()

        for index in closed:
            index, is_created = Index.objects.get_or_create(name=index)
            if str(index) in es_settings:
                unix_timestamp = int(es_settings[str(index)]['settings']['index']['creation_date']) / 1000
                utc_time = datetime.utcfromtimestamp(unix_timestamp).isoformat()
            index.created_at = utc_time
            index.save()
            if index.is_open:
                index.is_open = False
                index.save()


    @elastic_connection
    def get_indices(self) -> Tuple[list, list]:
        """
        Returns a tuple of open and closed list of indices that matches the ES_PREFIX,
        if it's not set, returns all indices in the server.
        """
        if self.connection:
            alias = '*'
            if self.es_prefix:
                alias = f'{self.es_prefix}*'
                opened = list(self.es.indices.get_alias(alias, expand_wildcards="open").keys())
                closed = list(self.es.indices.get_alias(alias, expand_wildcards="closed").keys())
                return opened, closed

            opened = list(self.es.indices.get_alias(alias, expand_wildcards="open").keys())
            closed = list(self.es.indices.get_alias(alias, expand_wildcards="closed").keys())
            return opened, closed
        else:
            raise NotFoundError(f"Error connecting to Elasticsearch: '{self.es_url}'! Do you have the right URL configured?")


    def handle_es6_mapping(self, mapping_schema: dict):
        properties = {}
        for doctype_key, mapping in mapping_schema.items():
            if mapping:
                mapping = mapping["properties"]
                for field_name, field_schema in mapping.items():
                    properties[field_name] = field_schema
        return properties


    @elastic_connection
    def get_fields(self, indices: List[str] = []):
        out = []
        indices = indices if indices else ["*"]
        if self.connection:
            for index in indices:
                mapping = Mapping.from_es(index=index, using=self.es)
                properties = {field: mapping[field].to_dict() for field in mapping}  # Only works for ES7 clusters.
                properties = properties if properties else self.handle_es6_mapping(mapping.to_dict())
                fields = self._decode_mapping_structure(properties)
                for field in fields:
                    index_with_field = {'index': index, 'path': field['path'], 'type': field['type']}
                    out.append(index_with_field)
        return out


    def _decode_mapping_structure(self, structure: dict, root_path=list()) -> List[dict]:
        """
        Decode mapping structure (nested dictionary) to a flat structure, separated by dot notation.

        :param structure: Dictionary where the keys are field names and their values their respective mapping in JSON format.
        :param root_path: Where to start from, used when dealing recursively with nested fields so the function would know where to start from.
        :return: List of dictionaries where the "path" key shows the dot notated path of a field and the "type" it's Elasticsearch data type.
        """
        mapping_data = []
        for k, v in structure.items():
            # deal with fact field
            if 'properties' in v and k in self.texta_reserved:
                sub_structure = v['properties']
                path_list = root_path[:]
                path_list.append(k)
                sub_mapping = [{'path': k, 'type': 'fact'}]
                mapping_data.extend(sub_mapping)
            # deal with object & nested structures
            elif 'properties' in v and k not in self.texta_reserved:
                sub_structure = v['properties']
                path_list = root_path[:]
                path_list.append(k)
                # detect MLP base fields in doc
                if k.endswith('_mlp'):
                    mlp_mapping = [{'path': '.'.join(path_list), 'type': 'mlp'}]
                    mapping_data.extend(mlp_mapping)
                # go deeper
                sub_mapping = self._decode_mapping_structure(sub_structure, root_path=path_list)
                mapping_data.extend(sub_mapping)
            # this is the end
            else:
                path_list = root_path[:]
                path_list.append(k)
                path = '.'.join(path_list)
                data = {'path': path, 'type': v['type']}
                mapping_data.append(data)
        return mapping_data


    @elastic_connection
    def check_if_indices_exist(self, indices: List[str]) -> bool:
        """
        Returns whether the given indices exist using a list of index
        names as input.
        """
        return self.es.indices.exists(index=','.join(indices))


    @elastic_connection
    def scroll(self, indices: List[str], query: dict, scroll_id: str = None, connection_timeout=60 * 1, scroll_timeout="10m", size=300, fields: List[str] = ["*"], with_meta=True):
        indices = ",".join(indices)
        if scroll_id is None:
            initial_scroll = self.es.search(index=indices, body=query, request_timeout=connection_timeout, scroll=scroll_timeout, size=size, _source=fields)
            documents = initial_scroll["hits"]["hits"] if with_meta else [doc["_source"] for doc in initial_scroll["hits"]["hits"]]
            total = initial_scroll["hits"]["total"]
            response = {
                "scroll_id": initial_scroll["_scroll_id"],
                "total_documents": total if isinstance(total, int) else total["value"],
                "returned_count": len(initial_scroll["hits"]["hits"]),
                "documents": documents
            }
            return response

        else:
            continuation_scroll = self.es.scroll(scroll_id=scroll_id, scroll=scroll_timeout)
            documents = continuation_scroll["hits"]["hits"] if with_meta else [doc["_source"] for doc in continuation_scroll["hits"]["hits"]]
            total = continuation_scroll["hits"]["total"]

            response = {
                "scroll_id": continuation_scroll["_scroll_id"],
                "total_documents": total if isinstance(total, int) else total["value"],
                "returned_count": len(continuation_scroll["hits"]["hits"]),
                "documents": documents
            }
            return response


    @staticmethod
    def parse_doc_type_from_mapping(mapping: dict, default_es7_doctype="_doc") -> str:
        """
        Parse the result of the indexes _mapping endpoint to fetch the set
        doc_type and if it doesn't exist yet (fresh index) put it into the default _doc.
        From the start of ES6, only a single doc_type is allowed per index.
        """
        doc_types = []

        for index, mappings in mapping.items():
            mappings = mappings["mappings"]
            for doc_type in mappings:
                if doc_type != "properties":
                    doc_types.append(doc_type)

        return doc_types[0] if doc_types else default_es7_doctype


    def get_doc_type_for_index(self, index: str):
        index_interface = elasticsearch_dsl.Index(index, using=self.es)
        mapping = index_interface.get_mapping()
        doc_type = self.parse_doc_type_from_mapping(mapping)
        return doc_type


    @elastic_connection
    def add_texta_facts_mapping(self, index: str):
        """
        To allow for more flexibility, we do not use the indices variable in the class.

        Adding the same mapping multiple times doesn't effect anything,
        adding a single field is also save as the query only adds, not overwrites.
        """
        m = Mapping()
        texta_facts = Nested(
            properties={
                "spans": Keyword(),
                "fact": Keyword(),
                "str_val": Keyword(),
                "doc_path": Keyword(),
                "num_val": Long(),
                "sent_index": Long()
            }
        )

        # Set the name of the field along with its mapping body
        mapping = m.field(TEXTA_TAGS_KEY, texta_facts).to_dict()
        doc_type = self.get_doc_type_for_index(index)
        self.es.indices.put_mapping(body=mapping, index=index, doc_type=doc_type, include_type_name=True)


    @elastic_connection
    def add_texta_meta_mapping(self, index: str):
        m = Mapping()
        texta_meta = Object(
            properties={
                "document_uuid": Text()
            }
        )

        # Set the name of the field along with its mapping body
        mapping = m.field("texta_meta", texta_meta).to_dict()
        doc_type = self.get_doc_type_for_index(index)
        self.es.indices.put_mapping(body=mapping, index=index, doc_type=doc_type, include_type_name=True)


    @elastic_connection
    def add_annotator_mapping(self, index: str):
        m = Mapping()
        texta_annotator = Nested(
            properties={
                "job_id": Long(),
                "processed_timestamp_utc": Date(),
                "skipped_timestamp_utc": Date(),
                "validated_timestamp_utc": Date(),
                "commented_timestamp_utc": Date()
            }
        )

        # Set the name of the field along with its mapping body
        mapping = m.field(TEXTA_ANNOTATOR_KEY, texta_annotator).to_dict()
        doc_type = self.get_doc_type_for_index(index)
        self.es.indices.put_mapping(body=mapping, index=index, doc_type=doc_type, include_type_name=True)


    def get_annotation_query(self, query: dict, job_pk: int):
        """
        Return a query dictionary for the annotator for documents that lack the given field
        within a limited subset.
        :param job_pk:
        :param query: Dictionary of an Elasticsearch query as an additional restriction.
        :return:
        """

        # TODO This can be written a bit better using elasticsearch_dsl query syntax to avoid pulling dicts from querys etc.
        negative_queries = [
            Q("match", **{f"{TEXTA_ANNOTATOR_KEY}.job_id": job_pk})
        ]
        positive_queries = [Q(query["query"])]
        search = elasticsearch_dsl.Search()
        restriction = Q("bool", must_not=negative_queries, must=positive_queries)
        search = search.query(restriction)
        return search.to_dict()


    def get_annotation_validation_query(self, query: dict):
        """
        Return a query dictionary for the annotator for documents that lack the given field
        within a limited subset for validation.
        :param query: Dictionary of an Elasticsearch query as an additional restriction.
        :return:
        """

        negative_queries = [Q("exists", field="texta_annotator.skipped_timestamp_utc"), Q("exists", field="texta_annotator.validated_timestamp_utc")]
        positive_queries = [Q(query["query"]), Q("exists", field="texta_annotator.processed_timestamp_utc")]
        s = Q("bool", must_not=negative_queries, must=positive_queries)
        return s.to_dict()


    def get_annotated_annotation_query(self, query: dict, job_pk: int):
        """
        Return a query dictionary for the annotator for documents that are already annotated.
        :param job_pk:
        :param query: Dictionary of an Elasticsearch query as an additional restriction.
        :return:
        """

        positive_queries = [
            Q(query["query"]),
            Q("match", **{f"{TEXTA_ANNOTATOR_KEY}.job_id": job_pk}),
            Q("exists", field=f"{TEXTA_ANNOTATOR_KEY}.processed_timestamp_utc")
        ]
        negative_queries = [
            Q("exists", field=f"{TEXTA_ANNOTATOR_KEY}.skipped_timestamp_utc")
        ]
        s = Q("bool", must_not=negative_queries, must=positive_queries)
        return s.to_dict()


    def get_skipped_annotation_query(self, query: dict, job_pk: int):
        """
        Return a query dictionary for the annotator for documents that lack the given field
        within a limited subset for validation.
        :param job_pk:
        :param query: Dictionary of an Elasticsearch query as an additional restriction.
        :return:
        """

        negative_queries = [
            Q("exists", field="texta_annotator.processed_timestamp_utc"),
            Q("exists", field="texta_annotator.validated_timestamp_utc")
        ]
        positive_queries = [
            Q(query["query"]),
            Q("match", **{f"{TEXTA_ANNOTATOR_KEY}.job_id": job_pk}),
            Q("exists", field=f"{TEXTA_ANNOTATOR_KEY}.skipped_timestamp_utc")
        ]
        s = Q("bool", must_not=negative_queries, must=positive_queries)
        return s.to_dict()


    def flatten(self, d, parent_key='', sep='.'):
        """
        From: https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys
        """
        items = []
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, MutableMapping):
                items.extend(self.flatten(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)


    @elastic_connection
    def get_index_stats(self, indices="_all"):
        store = {}

        # Get size of indices.
        response = self.es.indices.stats(index=indices, metric="store")
        for index_name in response["indices"].keys():
            size = response["indices"][index_name]["total"]["store"]["size_in_bytes"]
            store[index_name] = {"size": size, "doc_count": 0}  # Initialize doc_count as zero and overwrite later.

        # Get count of indices.
        s = elasticsearch_dsl.Search(using=self.es, index=indices).extra(size=0)
        s.aggs.bucket("by_all", "terms", field="_index", size=10000)
        for hit in s.execute().aggs.by_all:
            index_name = hit["key"]
            store[index_name]["doc_count"] = hit["doc_count"]

        return store
